@@ -28,7 +28,8 @@
 #include <grpcpp/impl/codegen/callback_common.h>
 #include <grpcpp/impl/codegen/config.h>
 #include <grpcpp/impl/codegen/core_codegen_interface.h>
-#include <grpcpp/impl/codegen/server_context.h>
+#include <grpcpp/impl/codegen/message_allocator.h>
+#include <grpcpp/impl/codegen/server_context_impl.h>
 #include <grpcpp/impl/codegen/server_interface.h>
 #include <grpcpp/impl/codegen/status.h>
 
@@ -37,11 +38,61 @@ namespace grpc {
 // Declare base class of all reactors as internal
 namespace internal {
 
+// Forward declarations
+template <class Request, class Response>
+class CallbackClientStreamingHandler;
+template <class Request, class Response>
+class CallbackServerStreamingHandler;
+template <class Request, class Response>
+class CallbackBidiHandler;
+
 class ServerReactor {
  public:
   virtual ~ServerReactor() = default;
-  virtual void OnDone() {}
-  virtual void OnCancel() {}
+  virtual void OnDone() = 0;
+  virtual void OnCancel() = 0;
+
+ private:
+  friend class ::grpc_impl::ServerContext;
+  template <class Request, class Response>
+  friend class CallbackClientStreamingHandler;
+  template <class Request, class Response>
+  friend class CallbackServerStreamingHandler;
+  template <class Request, class Response>
+  friend class CallbackBidiHandler;
+
+  // The ServerReactor is responsible for tracking when it is safe to call
+  // OnCancel. This function should not be called until after OnStarted is done
+  // and the RPC has completed with a cancellation. This is tracked by counting
+  // how many of these conditions have been met and calling OnCancel when none
+  // remain unmet.
+
+  void MaybeCallOnCancel() {
+    if (on_cancel_conditions_remaining_.fetch_sub(
+            1, std::memory_order_acq_rel) == 1) {
+      OnCancel();
+    }
+  }
+
+  std::atomic_int on_cancel_conditions_remaining_{2};
+};
+
+template <class Request, class Response>
+class DefaultMessageHolder
+    : public experimental::MessageHolder<Request, Response> {
+ public:
+  DefaultMessageHolder() {
+    this->set_request(&request_obj_);
+    this->set_response(&response_obj_);
+  }
+  void Release() override {
+    // the object is allocated in the call arena.
+    this->~DefaultMessageHolder<Request, Response>();
+  }
+
+ private:
+  Request request_obj_;
+  Response response_obj_;
 };
 
 }  // namespace internal
@@ -69,6 +120,44 @@ class ServerCallbackRpcController {
   // Allow the method handler to push out the initial metadata before
   // the response and status are ready
   virtual void SendInitialMetadata(std::function<void(bool)>) = 0;
+
+  /// SetCancelCallback passes in a callback to be called when the RPC is
+  /// canceled for whatever reason (streaming calls have OnCancel instead). This
+  /// is an advanced and uncommon use with several important restrictions. This
+  /// function may not be called more than once on the same RPC.
+  ///
+  /// If code calls SetCancelCallback on an RPC, it must also call
+  /// ClearCancelCallback before calling Finish on the RPC controller. That
+  /// method makes sure that no cancellation callback is executed for this RPC
+  /// beyond the point of its return. ClearCancelCallback may be called even if
+  /// SetCancelCallback was not called for this RPC, and it may be called
+  /// multiple times. It _must_ be called if SetCancelCallback was called for
+  /// this RPC.
+  ///
+  /// The callback should generally be lightweight and nonblocking and primarily
+  /// concerned with clearing application state related to the RPC or causing
+  /// operations (such as cancellations) to happen on dependent RPCs.
+  ///
+  /// If the RPC is already canceled at the time that SetCancelCallback is
+  /// called, the callback is invoked immediately.
+  ///
+  /// The cancellation callback may be executed concurrently with the method
+  /// handler that invokes it but will certainly not issue or execute after the
+  /// return of ClearCancelCallback. If ClearCancelCallback is invoked while the
+  /// callback is already executing, the callback will complete its execution
+  /// before ClearCancelCallback takes effect.
+  ///
+  /// To preserve the orderings described above, the callback may be called
+  /// under a lock that is also used for ClearCancelCallback and
+  /// ServerContext::IsCancelled, so the callback CANNOT call either of those
+  /// operations on this RPC or any other function that causes those operations
+  /// to be called before the callback completes.
+  virtual void SetCancelCallback(std::function<void()> callback) = 0;
+  virtual void ClearCancelCallback() = 0;
+
+  // NOTE: This is an API for advanced users who need custom allocators.
+  // Get and maybe mutate the allocator state associated with the current RPC.
+  virtual RpcAllocatorState* GetRpcAllocatorState() = 0;
 };
 
 // NOTE: The actual streaming object classes are provided
@@ -102,7 +191,7 @@ class ServerCallbackWriter {
     // Default implementation that can/should be overridden
     Write(msg, std::move(options));
     Finish(std::move(s));
-  };
+  }
 
  protected:
   template <class Request>
@@ -125,7 +214,7 @@ class ServerCallbackReaderWriter {
     // Default implementation that can/should be overridden
     Write(msg, std::move(options));
     Finish(std::move(s));
-  };
+  }
 
  protected:
   void BindReactor(ServerBidiReactor<Request, Response>* reactor) {
@@ -133,86 +222,214 @@ class ServerCallbackReaderWriter {
   }
 };
 
-// The following classes are reactors that are to be implemented
-// by the user, returned as the result of the method handler for
-// a callback method, and activated by the call to OnStarted
+// The following classes are the reactor interfaces that are to be implemented
+// by the user, returned as the result of the method handler for a callback
+// method, and activated by the call to OnStarted. The library guarantees that
+// OnStarted will be called for any reactor that has been created using a
+// method handler registered on a service. No operation initiation method may be
+// called until after the call to OnStarted.
+// Note that none of the classes are pure; all reactions have a default empty
+// reaction so that the user class only needs to override those classes that it
+// cares about.
+
+/// \a ServerBidiReactor is the interface for a bidirectional streaming RPC.
 template <class Request, class Response>
 class ServerBidiReactor : public internal::ServerReactor {
  public:
   ~ServerBidiReactor() = default;
-  virtual void OnStarted(ServerContext*) {}
+
+  /// Do NOT call any operation initiation method (names that start with Start)
+  /// until after the library has called OnStarted on this object.
+
+  /// Send any initial metadata stored in the RPC context. If not invoked,
+  /// any initial metadata will be passed along with the first Write or the
+  /// Finish (if there are no writes).
+  void StartSendInitialMetadata() { stream_->SendInitialMetadata(); }
+
+  /// Initiate a read operation.
+  ///
+  /// \param[out] req Where to eventually store the read message. Valid when
+  ///                 the library calls OnReadDone
+  void StartRead(Request* req) { stream_->Read(req); }
+
+  /// Initiate a write operation.
+  ///
+  /// \param[in] resp The message to be written. The library takes temporary
+  ///                 ownership until OnWriteDone, at which point the
+  ///                 application regains ownership of resp.
+  void StartWrite(const Response* resp) { StartWrite(resp, WriteOptions()); }
+
+  /// Initiate a write operation with specified options.
+  ///
+  /// \param[in] resp The message to be written. The library takes temporary
+  ///                 ownership until OnWriteDone, at which point the
+  ///                 application regains ownership of resp.
+  /// \param[in] options The WriteOptions to use for writing this message
+  void StartWrite(const Response* resp, WriteOptions options) {
+    stream_->Write(resp, std::move(options));
+  }
+
+  /// Initiate a write operation with specified options and final RPC Status,
+  /// which also causes any trailing metadata for this RPC to be sent out.
+  /// StartWriteAndFinish is like merging StartWriteLast and Finish into a
+  /// single step. A key difference, though, is that this operation doesn't have
+  /// an OnWriteDone reaction - it is considered complete only when OnDone is
+  /// available. An RPC can either have StartWriteAndFinish or Finish, but not
+  /// both.
+  ///
+  /// \param[in] resp The message to be written. The library takes temporary
+  ///                 ownership until Onone, at which point the application
+  ///                 regains ownership of resp.
+  /// \param[in] options The WriteOptions to use for writing this message
+  /// \param[in] s The status outcome of this RPC
+  void StartWriteAndFinish(const Response* resp, WriteOptions options,
+                           Status s) {
+    stream_->WriteAndFinish(resp, std::move(options), std::move(s));
+  }
+
+  /// Inform system of a planned write operation with specified options, but
+  /// allow the library to schedule the actual write coalesced with the writing
+  /// of trailing metadata (which takes place on a Finish call).
+  ///
+  /// \param[in] resp The message to be written. The library takes temporary
+  ///                 ownership until OnWriteDone, at which point the
+  ///                 application regains ownership of resp.
+  /// \param[in] options The WriteOptions to use for writing this message
+  void StartWriteLast(const Response* resp, WriteOptions options) {
+    StartWrite(resp, std::move(options.set_last_message()));
+  }
+
+  /// Indicate that the stream is to be finished and the trailing metadata and
+  /// RPC status are to be sent. Every RPC MUST be finished using either Finish
+  /// or StartWriteAndFinish (but not both), even if the RPC is already
+  /// cancelled.
+  ///
+  /// \param[in] s The status outcome of this RPC
+  void Finish(Status s) { stream_->Finish(std::move(s)); }
+
+  /// Notify the application that a streaming RPC has started and that it is now
+  /// ok to call any operation initiation method. An RPC is considered started
+  /// after the server has received all initial metadata from the client, which
+  /// is a result of the client calling StartCall().
+  ///
+  /// \param[in] context The context object now associated with this RPC
+  virtual void OnStarted(::grpc_impl::ServerContext* context) {}
+
+  /// Notifies the application that an explicit StartSendInitialMetadata
+  /// operation completed. Not used when the sending of initial metadata
+  /// piggybacks onto the first write.
+  ///
+  /// \param[in] ok Was it successful? If false, no further write-side operation
+  ///               will succeed.
   virtual void OnSendInitialMetadataDone(bool ok) {}
+
+  /// Notifies the application that a StartRead operation completed.
+  ///
+  /// \param[in] ok Was it successful? If false, no further read-side operation
+  ///               will succeed.
   virtual void OnReadDone(bool ok) {}
+
+  /// Notifies the application that a StartWrite (or StartWriteLast) operation
+  /// completed.
+  ///
+  /// \param[in] ok Was it successful? If false, no further write-side operation
+  ///               will succeed.
   virtual void OnWriteDone(bool ok) {}
 
-  void StartSendInitialMetadata() { stream_->SendInitialMetadata(); }
-  void StartRead(Request* msg) { stream_->Read(msg); }
-  void StartWrite(const Response* msg) { StartWrite(msg, WriteOptions()); }
-  void StartWrite(const Response* msg, WriteOptions options) {
-    stream_->Write(msg, std::move(options));
-  }
-  void StartWriteAndFinish(const Response* msg, WriteOptions options,
-                           Status s) {
-    stream_->WriteAndFinish(msg, std::move(options), std::move(s));
-  }
-  void StartWriteLast(const Response* msg, WriteOptions options) {
-    StartWrite(msg, std::move(options.set_last_message()));
-  }
-  void Finish(Status s) { stream_->Finish(std::move(s)); }
+  /// Notifies the application that all operations associated with this RPC
+  /// have completed. This is an override (from the internal base class) but not
+  /// final, so derived classes should override it if they want to take action.
+  void OnDone() override {}
+
+  /// Notifies the application that this RPC has been cancelled. This is an
+  /// override (from the internal base class) but not final, so derived classes
+  /// should override it if they want to take action.
+  void OnCancel() override {}
 
  private:
   friend class ServerCallbackReaderWriter<Request, Response>;
-  void BindStream(ServerCallbackReaderWriter<Request, Response>* stream) {
+  virtual void BindStream(
+      ServerCallbackReaderWriter<Request, Response>* stream) {
     stream_ = stream;
   }
 
   ServerCallbackReaderWriter<Request, Response>* stream_;
 };
 
+/// \a ServerReadReactor is the interface for a client-streaming RPC.
 template <class Request, class Response>
 class ServerReadReactor : public internal::ServerReactor {
  public:
   ~ServerReadReactor() = default;
-  virtual void OnStarted(ServerContext*, Response* resp) {}
+
+  /// The following operation initiations are exactly like ServerBidiReactor.
+  void StartSendInitialMetadata() { reader_->SendInitialMetadata(); }
+  void StartRead(Request* req) { reader_->Read(req); }
+  void Finish(Status s) { reader_->Finish(std::move(s)); }
+
+  /// Similar to ServerBidiReactor::OnStarted, except that this also provides
+  /// the response object that the stream fills in before calling Finish.
+  /// (It must be filled in if status is OK, but it may be filled in otherwise.)
+  ///
+  /// \param[in] context The context object now associated with this RPC
+  /// \param[in] resp The response object to be used by this RPC
+  virtual void OnStarted(::grpc_impl::ServerContext* context, Response* resp) {}
+
+  /// The following notifications are exactly like ServerBidiReactor.
   virtual void OnSendInitialMetadataDone(bool ok) {}
   virtual void OnReadDone(bool ok) {}
-
-  void StartSendInitialMetadata() { reader_->SendInitialMetadata(); }
-  void StartRead(Request* msg) { reader_->Read(msg); }
-  void Finish(Status s) { reader_->Finish(std::move(s)); }
+  void OnDone() override {}
+  void OnCancel() override {}
 
  private:
   friend class ServerCallbackReader<Request>;
-  void BindReader(ServerCallbackReader<Request>* reader) { reader_ = reader; }
+  virtual void BindReader(ServerCallbackReader<Request>* reader) {
+    reader_ = reader;
+  }
 
   ServerCallbackReader<Request>* reader_;
 };
 
+/// \a ServerWriteReactor is the interface for a server-streaming RPC.
 template <class Request, class Response>
 class ServerWriteReactor : public internal::ServerReactor {
  public:
   ~ServerWriteReactor() = default;
-  virtual void OnStarted(ServerContext*, const Request* req) {}
-  virtual void OnSendInitialMetadataDone(bool ok) {}
-  virtual void OnWriteDone(bool ok) {}
 
+  /// The following operation initiations are exactly like ServerBidiReactor.
   void StartSendInitialMetadata() { writer_->SendInitialMetadata(); }
-  void StartWrite(const Response* msg) { StartWrite(msg, WriteOptions()); }
-  void StartWrite(const Response* msg, WriteOptions options) {
-    writer_->Write(msg, std::move(options));
+  void StartWrite(const Response* resp) { StartWrite(resp, WriteOptions()); }
+  void StartWrite(const Response* resp, WriteOptions options) {
+    writer_->Write(resp, std::move(options));
   }
-  void StartWriteAndFinish(const Response* msg, WriteOptions options,
+  void StartWriteAndFinish(const Response* resp, WriteOptions options,
                            Status s) {
-    writer_->WriteAndFinish(msg, std::move(options), std::move(s));
+    writer_->WriteAndFinish(resp, std::move(options), std::move(s));
   }
-  void StartWriteLast(const Response* msg, WriteOptions options) {
-    StartWrite(msg, std::move(options.set_last_message()));
+  void StartWriteLast(const Response* resp, WriteOptions options) {
+    StartWrite(resp, std::move(options.set_last_message()));
   }
   void Finish(Status s) { writer_->Finish(std::move(s)); }
 
+  /// Similar to ServerBidiReactor::OnStarted, except that this also provides
+  /// the request object sent by the client.
+  ///
+  /// \param[in] context The context object now associated with this RPC
+  /// \param[in] req The request object sent by the client
+  virtual void OnStarted(::grpc_impl::ServerContext* context,
+                         const Request* req) {}
+
+  /// The following notifications are exactly like ServerBidiReactor.
+  virtual void OnSendInitialMetadataDone(bool ok) {}
+  virtual void OnWriteDone(bool ok) {}
+  void OnDone() override {}
+  void OnCancel() override {}
+
  private:
   friend class ServerCallbackWriter<Response>;
-  void BindWriter(ServerCallbackWriter<Response>* writer) { writer_ = writer; }
+  virtual void BindWriter(ServerCallbackWriter<Response>* writer) {
+    writer_ = writer;
+  }
 
   ServerCallbackWriter<Response>* writer_;
 };
@@ -226,7 +443,7 @@ class UnimplementedReadReactor
     : public experimental::ServerReadReactor<Request, Response> {
  public:
   void OnDone() override { delete this; }
-  void OnStarted(ServerContext*, Response*) override {
+  void OnStarted(::grpc_impl::ServerContext*, Response*) override {
     this->Finish(Status(StatusCode::UNIMPLEMENTED, ""));
   }
 };
@@ -236,7 +453,7 @@ class UnimplementedWriteReactor
     : public experimental::ServerWriteReactor<Request, Response> {
  public:
   void OnDone() override { delete this; }
-  void OnStarted(ServerContext*, const Request*) override {
+  void OnStarted(::grpc_impl::ServerContext*, const Request*) override {
     this->Finish(Status(StatusCode::UNIMPLEMENTED, ""));
   }
 };
@@ -246,7 +463,7 @@ class UnimplementedBidiReactor
     : public experimental::ServerBidiReactor<Request, Response> {
  public:
   void OnDone() override { delete this; }
-  void OnStarted(ServerContext*) override {
+  void OnStarted(::grpc_impl::ServerContext*) override {
     this->Finish(Status(StatusCode::UNIMPLEMENTED, ""));
   }
 };
@@ -255,21 +472,29 @@ template <class RequestType, class ResponseType>
 class CallbackUnaryHandler : public MethodHandler {
  public:
   CallbackUnaryHandler(
-      std::function<void(ServerContext*, const RequestType*, ResponseType*,
+      std::function<void(::grpc_impl::ServerContext*, const RequestType*,
+                         ResponseType*,
                          experimental::ServerCallbackRpcController*)>
           func)
       : func_(func) {}
+
+  void SetMessageAllocator(
+      experimental::MessageAllocator<RequestType, ResponseType>* allocator) {
+    allocator_ = allocator;
+  }
+
   void RunHandler(const HandlerParameter& param) final {
     // Arena allocate a controller structure (that includes request/response)
     g_core_codegen_interface->grpc_call_ref(param.call->call());
+    auto* allocator_state =
+        static_cast<experimental::MessageHolder<RequestType, ResponseType>*>(
+            param.internal_data);
     auto* controller = new (g_core_codegen_interface->grpc_call_arena_alloc(
         param.call->call(), sizeof(ServerCallbackRpcControllerImpl)))
-        ServerCallbackRpcControllerImpl(
-            param.server_context, param.call,
-            static_cast<RequestType*>(param.request),
-            std::move(param.call_requester));
+        ServerCallbackRpcControllerImpl(param.server_context, param.call,
+                                        allocator_state,
+                                        std::move(param.call_requester));
     Status status = param.status;
-
     if (status.ok()) {
       // Call the actual function handler and expect the user to call finish
       CatchingCallback(func_, param.server_context, controller->request(),
@@ -280,25 +505,38 @@ class CallbackUnaryHandler : public MethodHandler {
     }
   }
 
-  void* Deserialize(grpc_call* call, grpc_byte_buffer* req,
-                    Status* status) final {
+  void* Deserialize(grpc_call* call, grpc_byte_buffer* req, Status* status,
+                    void** handler_data) final {
     ByteBuffer buf;
     buf.set_buffer(req);
-    auto* request = new (g_core_codegen_interface->grpc_call_arena_alloc(
-        call, sizeof(RequestType))) RequestType();
+    RequestType* request = nullptr;
+    experimental::MessageHolder<RequestType, ResponseType>* allocator_state =
+        nullptr;
+    if (allocator_ != nullptr) {
+      allocator_state = allocator_->AllocateMessages();
+    } else {
+      allocator_state = new (g_core_codegen_interface->grpc_call_arena_alloc(
+          call, sizeof(DefaultMessageHolder<RequestType, ResponseType>)))
+          DefaultMessageHolder<RequestType, ResponseType>();
+    }
+    *handler_data = allocator_state;
+    request = allocator_state->request();
     *status = SerializationTraits<RequestType>::Deserialize(&buf, request);
     buf.Release();
     if (status->ok()) {
       return request;
     }
-    request->~RequestType();
+    // Clean up on deserialization failure.
+    allocator_state->Release();
     return nullptr;
   }
 
  private:
-  std::function<void(ServerContext*, const RequestType*, ResponseType*,
-                     experimental::ServerCallbackRpcController*)>
+  std::function<void(::grpc_impl::ServerContext*, const RequestType*,
+                     ResponseType*, experimental::ServerCallbackRpcController*)>
       func_;
+  experimental::MessageAllocator<RequestType, ResponseType>* allocator_ =
+      nullptr;
 
   // The implementation class of ServerCallbackRpcController is a private member
   // of CallbackUnaryHandler since it is never exposed anywhere, and this allows
@@ -320,7 +558,7 @@ class CallbackUnaryHandler : public MethodHandler {
       // The response is dropped if the status is not OK.
       if (s.ok()) {
         finish_ops_.ServerSendStatus(&ctx_->trailing_metadata_,
-                                     finish_ops_.SendMessagePtr(&resp_));
+                                     finish_ops_.SendMessagePtr(response()));
       } else {
         finish_ops_.ServerSendStatus(&ctx_->trailing_metadata_, s);
       }
@@ -349,28 +587,41 @@ class CallbackUnaryHandler : public MethodHandler {
       call_.PerformOps(&meta_ops_);
     }
 
+    // Neither SetCancelCallback nor ClearCancelCallback should affect the
+    // callbacks_outstanding_ count since they are paired and both must precede
+    // the invocation of Finish (if they are used at all)
+    void SetCancelCallback(std::function<void()> callback) override {
+      ctx_->SetCancelCallback(std::move(callback));
+    }
+
+    void ClearCancelCallback() override { ctx_->ClearCancelCallback(); }
+
+    experimental::RpcAllocatorState* GetRpcAllocatorState() override {
+      return allocator_state_;
+    }
+
    private:
     friend class CallbackUnaryHandler<RequestType, ResponseType>;
 
-    ServerCallbackRpcControllerImpl(ServerContext* ctx, Call* call,
-                                    const RequestType* req,
-                                    std::function<void()> call_requester)
+    ServerCallbackRpcControllerImpl(
+        ::grpc_impl::ServerContext* ctx, Call* call,
+        experimental::MessageHolder<RequestType, ResponseType>* allocator_state,
+        std::function<void()> call_requester)
         : ctx_(ctx),
           call_(*call),
-          req_(req),
+          allocator_state_(allocator_state),
           call_requester_(std::move(call_requester)) {
       ctx_->BeginCompletionOp(call, [this](bool) { MaybeDone(); }, nullptr);
     }
 
-    ~ServerCallbackRpcControllerImpl() { req_->~RequestType(); }
-
-    const RequestType* request() { return req_; }
-    ResponseType* response() { return &resp_; }
+    const RequestType* request() { return allocator_state_->request(); }
+    ResponseType* response() { return allocator_state_->response(); }
 
     void MaybeDone() {
       if (--callbacks_outstanding_ == 0) {
         grpc_call* call = call_.call();
         auto call_requester = std::move(call_requester_);
+        allocator_state_->Release();
         this->~ServerCallbackRpcControllerImpl();  // explicitly call destructor
         g_core_codegen_interface->grpc_call_unref(call);
         call_requester();
@@ -384,10 +635,10 @@ class CallbackUnaryHandler : public MethodHandler {
         finish_ops_;
     CallbackWithSuccessTag finish_tag_;
 
-    ServerContext* ctx_;
+    ::grpc_impl::ServerContext* ctx_;
     Call call_;
-    const RequestType* req_;
-    ResponseType resp_;
+    experimental::MessageHolder<RequestType, ResponseType>* const
+        allocator_state_;
     std::function<void()> call_requester_;
     std::atomic_int callbacks_outstanding_{
         2};  // reserve for Finish and CompletionOp
@@ -425,6 +676,8 @@ class CallbackClientStreamingHandler : public MethodHandler {
 
     reader->BindReactor(reactor);
     reactor->OnStarted(param.server_context, reader->response());
+    // The earliest that OnCancel can be called is after OnStarted is done.
+    reactor->MaybeCallOnCancel();
     reader->MaybeDone();
   }
 
@@ -486,7 +739,8 @@ class CallbackClientStreamingHandler : public MethodHandler {
     friend class CallbackClientStreamingHandler<RequestType, ResponseType>;
 
     ServerCallbackReaderImpl(
-        ServerContext* ctx, Call* call, std::function<void()> call_requester,
+        ::grpc_impl::ServerContext* ctx, Call* call,
+        std::function<void()> call_requester,
         experimental::ServerReadReactor<RequestType, ResponseType>* reactor)
         : ctx_(ctx),
           call_(*call),
@@ -526,7 +780,7 @@ class CallbackClientStreamingHandler : public MethodHandler {
     CallOpSet<CallOpRecvMessage<RequestType>> read_ops_;
     CallbackWithSuccessTag read_tag_;
 
-    ServerContext* ctx_;
+    ::grpc_impl::ServerContext* ctx_;
     Call call_;
     ResponseType resp_;
     std::function<void()> call_requester_;
@@ -567,11 +821,13 @@ class CallbackServerStreamingHandler : public MethodHandler {
                                  std::move(param.call_requester), reactor);
     writer->BindReactor(reactor);
     reactor->OnStarted(param.server_context, writer->request());
+    // The earliest that OnCancel can be called is after OnStarted is done.
+    reactor->MaybeCallOnCancel();
     writer->MaybeDone();
   }
 
-  void* Deserialize(grpc_call* call, grpc_byte_buffer* req,
-                    Status* status) final {
+  void* Deserialize(grpc_call* call, grpc_byte_buffer* req, Status* status,
+                    void** handler_data) final {
     ByteBuffer buf;
     buf.set_buffer(req);
     auto* request = new (g_core_codegen_interface->grpc_call_arena_alloc(
@@ -661,7 +917,7 @@ class CallbackServerStreamingHandler : public MethodHandler {
     friend class CallbackServerStreamingHandler<RequestType, ResponseType>;
 
     ServerCallbackWriterImpl(
-        ServerContext* ctx, Call* call, const RequestType* req,
+        ::grpc_impl::ServerContext* ctx, Call* call, const RequestType* req,
         std::function<void()> call_requester,
         experimental::ServerWriteReactor<RequestType, ResponseType>* reactor)
         : ctx_(ctx),
@@ -702,7 +958,7 @@ class CallbackServerStreamingHandler : public MethodHandler {
     CallOpSet<CallOpSendInitialMetadata, CallOpSendMessage> write_ops_;
     CallbackWithSuccessTag write_tag_;
 
-    ServerContext* ctx_;
+    ::grpc_impl::ServerContext* ctx_;
     Call call_;
     const RequestType* req_;
     std::function<void()> call_requester_;
@@ -743,6 +999,8 @@ class CallbackBidiHandler : public MethodHandler {
 
     stream->BindReactor(reactor);
     reactor->OnStarted(param.server_context);
+    // The earliest that OnCancel can be called is after OnStarted is done.
+    reactor->MaybeCallOnCancel();
     stream->MaybeDone();
   }
 
@@ -828,7 +1086,8 @@ class CallbackBidiHandler : public MethodHandler {
     friend class CallbackBidiHandler<RequestType, ResponseType>;
 
     ServerCallbackReaderWriterImpl(
-        ServerContext* ctx, Call* call, std::function<void()> call_requester,
+        ::grpc_impl::ServerContext* ctx, Call* call,
+        std::function<void()> call_requester,
         experimental::ServerBidiReactor<RequestType, ResponseType>* reactor)
         : ctx_(ctx),
           call_(*call),
@@ -874,7 +1133,7 @@ class CallbackBidiHandler : public MethodHandler {
     CallOpSet<CallOpRecvMessage<RequestType>> read_ops_;
     CallbackWithSuccessTag read_tag_;
 
-    ServerContext* ctx_;
+    ::grpc_impl::ServerContext* ctx_;
     Call call_;
     std::function<void()> call_requester_;
     experimental::ServerBidiReactor<RequestType, ResponseType>* reactor_;
